'''
base 0.4.2
Basic scanners for sagator.

(c) 2003 Jan ONDREJ (SAL) <ondrejj@salstar.sk>

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.
    
'''

from avlib import *
import re,string,os

__all__=['regexp_scan','string_scan',
         'file2stream','stream2file','f2s','s2f',
         'l_return',
         'match_all','alternatives','match_any',
         'max_file_size','exec_any']

class regexp_scan(ascanner):
  '''
  Primitive regexp pattern scanner.
  
  Usage: regexp_scan({'VirName':['RegExp_Pattern...',...],...} [,size])

  Where: 'VirName' is a string, which identifies defined virus
         'RegExp_Pattern...' is a regexp pattern
         size is a number, which defines, how many bytes may be checked.
           If it is not defined, whole stream is scanned.
    
  There can be more patterns for one virus. All patterns
  must match to assign an stream as virus.
  There can also be more virnames in one dictionary.
  '''
  name='RegExpScanner()'
  def __init__(self,regexps,size=0):
      self.regexps=regexps
      self.size=size
  def scanstream(self,stream):
      for vname,regs in self.regexps.items():
        for reg in regs:
          if self.size>0:
            if re.search(reg,stream[:self.size]):
              if reg==regs[-1]:
                return vname,['regexp: '+vname+' FOUND, '+reg+'\n']
              else:
                break
          else:
            if re.search(reg,stream):
              if reg==regs[-1]:
                return vname,['regexp: '+vname+' FOUND, '+reg+'\n']
            else:
              break
      return '',[]

class string_scan(ascanner):
  '''
  Primitive string pattern scanner.
  
  Usage: string_scan({'VirName':['Pattern...',...],...} [,size])

  Where: 'VirName' is a string, which identifies defined virus
         'Pattern...' is a string pattern
         size is a number, which defines, how many bytes may be checked.
           If it is not defined, whole stream is scanned.
    
  There can be more patterns for one virus. All patterns
  must match to assign an stream as virus.
  There can also be more virnames in one dictionary.
  '''
  name='StringScanner()'
  def __init__(self,strings,size=0):
      self.strings=strings
      self.size=size
  def scanstream(self,stream):
      for vname,regs in self.strings.items():
        for reg in regs:
          if self.size>0:
            if string.find(stream[:self.size],reg)>=0:
              if reg==regs[-1]:
                return vname,['string: '+vname+' FOUND, '+reg+'\n']
            else:
              break
          else:
            if string.find(stream,reg)>=0:
              if reg==regs[-1]:
                return vname,['string: '+vname+' FOUND, '+reg+'\n']
            else:
              break
      return '',[]

class file2stream(interscanner):
  '''
  file2stream interscanner
  
  Usage: file2stream(scanner())
  
  Where: scanner() is an interscanner or realscanner. It must be a streamscanner.
  
  This scanner converts a filescanner input into streamscanner.
  '''
  name='file2stream()'
  def __init__(self,scanner):
      self.name=self.name[:-1]+scanner.name+')'
      self.scanner=scanner
  def scanfile(self,files,dir=''):
      for fn in files:
        f=open(fn)
        vir,ret=self.scanner.scanstream(f.read())
        f.close()
        if vir:
          return vir,ret
      return '',[]

class f2s(file2stream):
  '''An alias for file2stream interscanner'''
  name='f2s()'

class stream2file(interscanner):
  '''
  stream2file interscanner
  
  Usage: stream2file(scanner())
  
  Where: scanner() is a interscanner or realscanner. It must be a filescanner.
  
  You can use it to save streams into files.
  '''
  name='stream2file()'
  is_multistream=1
  def __init__(self,scanner,addheader=0):
      self.name=self.name[:-1]+scanner.name+')'
      self.scanner=scanner
      self.addheader=addheader
      self.filename=''
      self.mboxdir=''
      self.filenames=[]
      self.counter=1
  def destroy(self):
      self.scanner.destroy()
      rdel([self.mboxdir])
      self.filenames=[]
      self.mboxdir=''
  def addstream(self,stream):
      if stream:
        if self.mboxdir=='':
          self.mbox=mktemp('/tmp/s2f-','.mbd','d',0700)
          self.mboxdir=self.mbox.name
        if self.filename=='':
          self.filename='part-%06d' % (self.counter)
          self.counter=self.counter+1
        mboxfile=mktemp(self.mboxdir+'/','-'+self.filename,'w',0600)
        creatembox(mboxfile.f,stream,self.addheader)
        self.filenames.append(mboxfile.name)
  def scanstream(self,stream):
      vir,ret='',[]
      self.addstream(stream)
      try:
        vir,ret=self.scanner.scanfile(self.filenames,self.mboxdir)
      except:
        raise
      return vir,ret
  def scanmultistream(self,streams):
      vir,ret='',[]
      if type(streams)==type({}):
        for filename,stream in streams.items():
          self.addstream(stream,filename)
      else:
        for stream in streams:
          self.addstream(stream)
      try:
        vir,ret=self.scanner.scanfile(self.filenames,self.mboxdir)
      except:
        raise
      return vir,ret

class s2f(stream2file):
  '''An alias for stream2file interscanner.'''
  name='s2f()'

class l_return(ascanner):
  '''
  l_return() realscanner
  
  Usage: l_return('VirName',['virreturnstringline1',...])
      or l_return()
  
  Where: 'VirName',[...] is a returned virus.
  
  This scanner has no real functionality. It return always
  a defined virus, or clean (if virus is not defined).
  This scanner has no error codes.
  '''
  name='l_return()'
  def __init__(self,vir='',ret=[]):
      self.vir=''+vir
      self.ret=[]+ret
      self.filename=''
  def scanstream(self,stream):
      if self.filename:
        debug(3,self.name,": Scanning file: ",self.filename)
      return self.vir,self.ret
  def scanfile(self,files,dirname=''):
      for file in files:
        debug(3,self.name,": Scanning file: ",file)
      return self.vir,self.ret

class match_class(interscanner):
  '''parent class for all match clases'''
  def __init__(self,*scanners):
      self.scanners=scanners
      n=[]
      for s in scanners:
        n=n+[s.name]
      self.name=self.name[:-1]+string.join(n,',')+')'
  def destroy(self):
      for scnr in self.scanners:
        scnr.destroy()
  def reinit(self):
      for scnr in self.scanners:
        scnr.reinit()

class match_all(match_class):
  '''
  match_all() interscanner
  
  Usage: match_all(scanner1(),scanner2(),...)
  
  Where: scanner1(),... are scanners, which must find a virus
  
  Returns a virus only if all scanners returns a virus.
  '''
  name='match_all()'
  def scanstream(self,stream):
      vir,ret='',[]
      for s in self.scanners:
        v,r=s.scanstream(stream)
        if v:
          if vir:
            vir=vir+','+v
          else:
            vir=v
          ret=ret+r
        else:
          return '',[]
      return vir,ret

class alternatives(match_class):
  '''
  alternatives() interscanner
  
  Usage: alternatives(scanner1(),scanner2(),...)
  
  Where: scanner1(),... are scanners, which may find a virus
  
  Returns a virus, if one scanner returns a virus.
  Scannig is ended if one scanner returns virus or clean.
  Next scanner from the list is used, if previous scannes fails
  (raises an error). Raises an ScannerError, if all scanners failed.
  '''
  name='alternatives()'
  def scanstream(self,stream):
      vir,ret='',[]
      for s in self.scanners:
        try:
          vir,ret=s.scanstream(stream)
          return vir,ret
        except:
          debug_traceback(4,'')
          debug(3,self.name,": scanner ",s.name," failed: ",
                string.rstrip(traceback_value_str()))
      debug(3,self.name,": All scanners failed")
      raise ScannerError,''

class match_any(match_class):
  '''
  match_any() interscanner
  
  Usage: match_any(scanner1(),scanner2(),...)
  
  Where: scanner1(),... are scanners, which may find a virus
  
  Returns a virus, if one scanner returns a virus.
  Next scanner from list is used, if previous scanner returns
  no virus, or failed (raised an error).
  '''
  name='match_any()'
  def scanstream(self,stream):
      vir,ret='',[]
      for s in self.scanners:
        try:
          vir,ret=s.scanstream(stream)
          if vir:
            return vir,ret
        except:
          debug_traceback(4,'')
          debug(3,self.name,": scanner ",s.name," failed: ",
                self.rstrip(traceback_value_str()))
      return '',[]

class max_file_size(ascanner):
  '''
  max_file_size() realscanner
  
  Usage: max_file_size(size[,'name'])
  
  Where: size is a number of bytes, which if exceeded, virus is returned
         'name' is a string, which identifies the virus name.
            If it is not present, 'FileSizeOverrun' is returned.
  '''
  name='max_file_size()'
  def __init__(self,size,vname='FileSizeOverrun'):
      self.size=size
      self.vname=vname
  def scanstream(self,stream):
      if len(stream)<=self.size:
        return '',[]
      else:
        return self.vname,[self.filename+": "+self.vname]
  def scanfile(self,files,dir=''):
      for file in files:
        if os.lstat(file)[6]>self.size:
          return self.vname,[file+': '+self.vname]
      return '',[]

class exec_any(ascanner):
  '''
  exec_any() realscanner
  
  Usage: exec_any(['command','arg1',...])
  
  Where: 'command' is a string, which command to run
         'argX' are command line arguments
  
  If command returns exit status 0, no virus is returned.
  Otherwise error is raised. This scanner is only for testing,
  or as development example.
  '''
  name='exec_any()'
  def __init__(self,arg):
      self.arg=arg
  def scanstream(self,stream):
      pf=popen(self.arg)
      pf.wait()
      pf.close()
      if pf.exitstatus:
        raise ScannerError,'EXITSTATUS: '+str(pf.exitstatus)
      return '',[]
  def scanfile(self,files,dir=''):
      return self.scanstream('')
